/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




#ifndef _CKT_INCLUDE_
#define _CKT_INCLUDE_


#include "DB/DB.h"
#include "PhySyn/libcell.h"
#include "PhySyn/utils.h"
#include "ABKCommon/uofm_alloc.h"

#include <map>
#include <fstream>
#include <bitset>

typedef std::map<uofm::string, class b_cell*> sym_map;

enum cell_type {INST, WIRE, PORT};

// Basic cell. Contains type only.
class b_cell
{
public:
  virtual ~b_cell() {}
  void set_name(const uofm::string& my_name)
  {
    name= my_name;
  }
  uofm::string& get_name()
  {
    return name;
  }
  virtual cell_type get_type() = 0;
  uofm::string name;
};

// Port
class port: public b_cell
{
public:
  port(): db_pin(NULL), fixed(false), redundant(false), arrival_time(0.) {}
  cell_type get_type()
  {
    return PORT;
  }
  void connect_wire(class wire* in_wire)
  {
    wire1= in_wire;
  }
  void set_inst(class inst* in_inst)
  {
    inst1= in_inst;
  }
  double get_HPWL(port* port2);
  void dump();
  uofm::string& get_fullname();
  uofm::string strbuf;
  // The wire it connects to
  wire* wire1;
  inst* inst1;
  dbPin* db_pin;
  double x, y;
  bool fixed; // Become true if it is moved, so cannot be moved again
  bool redundant; // Become true if it is determined to be redundant
  double arrival_time;
  double require_time;
  double slack;
  bool flag1; // temporary flag
};

// Instance
class inst: public b_cell
{
public:
  enum direction_t {N, S, E, W, FN, FS, FE, FW};

  inst(): x(0), y(0), is_latch(false), is_port(false), visited(false), sta_visited(false),
      from_bus(false), direction(N) {}
  cell_type get_type()
  {
    return INST;
  }
  void evaluate();
  void evaluate_core();
  void add_input(port* in_port)
  {
    inputs.push_back(in_port);
  }
  void add_output(port* in_port)
  {
    outputs.push_back(in_port);
  }
  void set_ttable(int input_no, const uofm::string& in_blif, libcell_set& lib_table);
  void set_ttable(int input_no, unsigned long tnumber, libcell_set& lib_table);
  void redirect();
  void unredirect();
  int get_func(int output_no); // return instance type
  uofm::string get_newname();
  int blif_match(const uofm::string& inputstr, const uofm::string& in_blif);
  void verify_evaluate_core();
  bool find_port_insert(const uofm::string& port_name, const uofm::string& io_direct, double x, double y, wire* wire2);
  void port_reorder(const uofm::vector<uofm::string>& inputstr, const uofm::vector<uofm::string>& outputstr,
                    uofm::vector<unsigned long>& ttables, libcell_set& lib_table);
  // Jarrod: Call this function after moving an instance(cell) and call
  // timing analyzer
  void move_to(double new_x, double new_y);
  int get_strength();
  void set_type_name(const uofm::string& inname);
  void dump();
  void transform_point(double x1, double y1, double& nx, double& ny);

  uofm::vector<lib_cell*> ttable; // truth table of the instance, points to lib_cell
  // each libcell contains one turth table
  uofm::string type_name; // Name of cell type

  // coordinates of the instance. It is the "center" of the instance
  double x, y;
  double sizex, sizey;
  bool is_latch;
  bool is_port;
  bool moved;
  bool relocated;
  bool visited;
  bool sta_visited;
  bool from_bus; // from terminal which is bus
  uofm::vector<port*> inputs;
  uofm::vector<port*> outputs;
  direction_t direction;
  double gate_delay;
  class inst *prev, *next; // linked list for STA, levelized prev and next
  double sta_order; // Order of evaluation for STA. Smaller number is earlier.
  // May be fraction.
  int level; // Level of the instance
  double criticality; // It is the maximum criticality of its fanout wires
};

// Wire
class wire: public b_cell
{
public:
  wire();
  cell_type get_type()
  {
    return WIRE;
  };
  void set_driver(port* in_port)
  {
    driver= in_port;
    wl_valid= false;
    in_port->wire1= this;
  };
  port* get_driver()
  {
    return driver;
  };
  double get_HPWL();
  double get_HPWL_execute();
  double get_HPWL_add(port* port1);
  double get_HPWL_remove(port* port1);
  void get_cog(double& x, double& y); // return center-of-gravity
  void get_cog_no_driver(double& x, double& y); // return center-of-gravity of fanouts
  void dump();
  void get_HPWL_remove_wire(double& wl, int level, uofm::vector<double>& dvecs);
  void add_port(port* port2, int isw);
  void delete_port(port* port2, int isw);
  uofm::vector<port*> outputs;
  int fanout;
  int cfanin, cfanout; // temp variable used in subckt extraction
  bool buf_tried; // if true, this wire has been buffered (or tried)
  port* driver;
  // For HPWL calculation
  double minx, miny, maxx, maxy;
  int value;
  unsigned long signature, sig_temp; // The pattern of wire from simulation
  bool restructured;
  bool wl_valid; // wirelength is valid. No need to recalculate
  bool is_clock;
  double hpwl; // HPWL
  uofm::vector<unsigned long> signatures;
  int one_count;
  uofm::vector<double> routed_wire_delay; // Saves wire delay from driver to each fanout
  double total_c; // total C of the RC tree
  bool rc_valid;
  bool route_valid; // true if routing information is still valid
  // (will be cleared after any change to the wire)
  double slack; // uses the slack of its driver
  double criticality;
};

// The circuit
class ckt
{
public:
  class Parameters
  {
  public:
    bool logic; // Uses bookshelf logic file
    bool pin_offset; // pin has offset
    bool real_size; // Uses real size (for blif2book)
    int mult_size_x; // X multiplier (for blif2book)
    int mult_size_y; // Y multiplier (for blif2book)
    bool mult_size; // Use multi_size (for blif2book)
    bool call_fraig; // Generate negative output latch for calling FRAIG
    Parameters();
  };

  ckt(Parameters params = Parameters()); // Build an empty circuit
  ckt(const char* aux_file_name, Parameters params = Parameters());
  ckt(DB* in_db, Parameters params = Parameters());
  ~ckt();
  int read_blif(const char* filename);
  int read_layout(const char* filename, int isw);
  int read_layout_new(const char* filename, int isw);
  int read_net(const char* filename);
  int read_net_new(const char* filename);
  void cname_to_name(uofm::string& name);
  void write_book(const char* filename);
  void write_nets(const char* filename);
  void write_nodes(const char* nodefilename);
  void write_logic(const char* logicfilename);
  b_cell* sym_table_find(uofm::string& name);
  // Get half perimeter wire length
  double get_HPWL();
  double get_HPWL_execute();
  void rewire(port* port2, port* driver);
  void rewire_drivers(uofm::vector<port*> orig_drivers, uofm::vector<port*>& new_drivers);
  void swap_driver(wire* wire1, wire* wire2);
  void swap_driver_port(wire* wire1, wire* wire2, uofm::vector<port*>& ports);
  void remove_redundant();
  void remove_port(port* port2);
  void write_blif(const char* filename);
  void write_pl(const char* filename);
  void remove_redundant_gate();
  void insert_inst(inst* inst2);
  double get_rewire_HPWL(port* port1, wire* wire1, wire* wire2);
  double get_wires_HPWL(uofm::vector<wire*>& wires);
  bool check_input_cone(port* port2, port* driver);
  void set_wire_value(int val);
  sym_map sym_table;
  libcell_set lib_table;
  int init();
  bool debug_remove_gate;
  void inst_bfsdfs(inst* inst2, int inst_no, uofm::vector<inst*>& insts, int isw);
  void inst_extract(inst* inst2, int inst_no, uofm::vector<inst*>& insts, int isw);
  void subckt_get_io(uofm::vector<inst*>& insts, uofm::vector<wire*>& iwires, uofm::vector<wire*>& owires, uofm::vector<port*>& inputs, uofm::vector<port*>& outputs);
  void rewire_input_wires(uofm::vector<wire*>& iwires, uofm::vector<wire*>& niwires, uofm::vector<port*>& inputs);
  void write_def(char* filename);
  void write_def_port(std::ofstream& ofile, wire* wire2, port* port2);
  void operator=(const ckt& in_ckt);
  void get_insts(uofm::vector<inst*>& insts);
  void get_wires(uofm::vector<wire*>& wires);
  void get_wires_sort_wl(uofm::vector<wire*>& wires);
  void get_insts_constraints(uofm::vector<inst*>& insts, double time2, double x2, double y2, double dist2);
  void generate_blif_for_lib();
  void clear_visited_flag();
  void call_fraig();
  b_cell* find_cell(const uofm::string& inname);
  void update_from_def(const char* filename);
  void print_move(inst* inst2);
  void dump();
  wire* DB_id_get_wire(int id)
  {
    return id_wire_vec[id];
  }
  inst* DB_id_get_inst(int id)
  {
    return id_inst_vec[id];
  }
  int pattern_no; // Number of patterns used in simulation
  DB *db;

private:
  uofm::vector<wire*> id_wire_vec;
  uofm::vector<inst*> id_inst_vec;
  void init_var();
  int parse_blif(const char* filename);
  int get_blif_token(uofm::string& token);
  int get_blif_ttable(uofm::string& token);
  int read_logic(const char* filename);
  wire* locate_wire_from_net(uofm::vector<uofm::string>& strvec);
  wire* find_wire_insert(uofm::string& name);
  bool wire_is_net(wire* wire2, uofm::vector<uofm::string>& strvec);
  std::fstream ifile;
  std::fstream fraig_aux;
  std::fstream inst_file; // File for gates inserted/deleted
  int inst_no, wire_no;
  int port_no;
  int term_no; // Number of inputs/outputs
  utils util;
  Parameters params1;
};

// Data structure for wire sorting
class sort_wire
{
public:
  double wl;
  wire* wire1;
  bool operator<(const sort_wire& s1) const
  {
    return wl < s1.wl;
  }
};

#endif
