/**************************************************************************
***
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2007 Regents of the University of Michigan,
***               Saurabh N. Adya, Jarrod A. Roy, David A. Papa and
***               Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation
***  the rights to use, copy, modify, merge, publish, distribute, sublicense,
***  and/or sell copies of the Software, and to permit persons to whom the
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/


#ifndef _STA_INCLUDE_
#define _STA_INCLUDE_

#include "Geoms/bbox.h"
#include "PhySyn/ckt.h"
#include "PhySyn/blif.h"
#include "TimeAna/rctree.h"

class sta
{
public:
  class Parameters
  {
  public:
    double buffer_delay; // the delay of buffer
    double portC; // Capacitance of a port
    double unitC; // C per unit
    double unitR; // R per unit
    double driverR; // R of a driver
    double intrinsic_delay; // intrinsic delay constant
    double database_microns; // From LEF. Conversion to real um.
    double gate_width_delay; // Used to increase delay according to gate width. Set it to 0.0 will make all gate delay identical (=intrinsic_delay)
    bool use_routed_wl; // Use routed wirelength from DB
    bool use_mst; // use MST: either FLUTE, MST_SIMPLE or Jarrod's supper FLUTE
    bool use_mst_simple; // Use my MST
    bool use_mst_super; // Use super flute
    bool use_flute; // Use FLUTE package (Steiner tree)
    bool ideal_input; // Use ideal input
    bool use_route_init; // Use routed delay in the beginning
    int plot; // plot critical paths before/after optimzation
    int plot2; // plot the same critical paths before/after optimization
    uofm::string DFF_D; // The keyword to recognize a DFF's D
    uofm::string DFF_CLK; // The keyword to recognize a DFF's CLK
    Parameters();
  };

  sta(ckt& inckt, Parameters params = Parameters());
  ckt& ckt1;
  ckt& get_internal_ckt()
  {
    return ckt1;
  }
  void init();
  void whole_chip(); // whole chip analysis
  void incremental(wire* iwire); // incremental analysis from wire
  void incremental(inst* iinst); // incremental analysis from instance
  void incremental_inst(inst* inst2); // incremental analysis from instance's inputs. Jarrod, you should use this
  double find_critical_path(uofm::vector<wire*>& wires);
  double get_max_delay();
  void write_plot(char* filename);
  void write_plot2(char* filename, uofm::vector<port*>& ports);
  void write_port_path(std::ofstream& ofile, port* iport);
  void get_critical_port(int n, uofm::vector<port*>& ports);
  void insert_inst(inst* inst2);
  void delete_inst(inst* inst2);
  void get_ordered_inst(inst* inst2, inst*& ninst, int isw);
  void get_output_cone(uofm::vector<inst*>& iinsts, uofm::vector<inst*>& insts);
  void get_output_cone(inst* inst2, uofm::vector<inst*>& insts);
  void clear_output_cone(uofm::vector<inst*>& insts);
  void incremental_insts(uofm::vector<inst*>& insts);
  // update criticality after timing changed.
  void update_criticality();
  void set_obstacles(uofm::vector<BBox> obstacle_boxes);
  bool debug;
  blif blif2;
  bool debug_incre; // debug incremental
  bool comp_on; // If true, compare time with calculated time
  uofm::vector<wire*> criticality_vec; // for wire
  uofm::vector<inst*> inst_criticality_vec; // for inst

private:
  void cal_slack();
  void levelize(uofm::vector<inst*>& sinsts);
  void cal_inst_delay(inst* inst2);
  void cal_wire_delay(wire* wire2);
  void cal_wire_req(wire* wire2);
  void cal_routed_wire_delay(wire* wire2);
  void cal_flute_wire_delay(wire* wire2);
  void build_llist_from_blif();
  double get_HPWL_two(double x1, double y1, double x2, double y2);
  void tree_cal_rc(rc_tree& in_tree);
  void tree_cal_delay(rc_tree& in_tree, wire* wire2, int m);
  uofm::vector<BBox> obstacles;
  inst* linst_head;  // the 1st instance of levelized instance list
  inst* linst_end;
  uofm::vector<port*> output_ports;
  Parameters params1;
};

class sort_wire_criticality
{
public:
  bool operator()(wire* w1, wire* w2)
  {
    return w1->slack < w2->slack;
  }
};

class sort_inst_criticality
{
public:
  bool operator()(inst* i1, inst* i2)
  {
    return i1->criticality > i2->criticality;
  }
};

#endif


